# -*- coding: utf-8 -*-
#______________________________________________________________________________
# General modules -------------------------------------------------------------
import numpy as np
import random as rnd

# Own modules -----------------------------------------------------------------
from molecule import VSSMolecule
from constants import kB,u
from AllSpecies import SpeciesDict
#______________________________________________________________________________

class DSMCSpecies(VSSMolecule):
    
    # Constructor -------------------------------------------------------------
    # Create an object based on the VSSMolecule of Podostemaceaous, Developed 
    # by J. Schwartz, that can be used as a variable soft sphere in DSMC simulations
    # with corresponding limits.
    #__________________________________________________________________________
    
    def __init__(self, name, mass, diameter, omega, t_ref, alpha):
        super().__init__(name, mass, diameter, omega=omega, t_ref=t_ref, alpha=alpha)
    
    # DSMCLimits --------------------------------------------------------------
    # Determine SPARTA DSMC simulation settings for a vapor cloud of specified
    # temperature and a given simulation domain. Accuracy demands can be changed
    # if necessary but basic parameters comply with Sun et al., 2011. The simulation
    # settings can be printed to the screen or returned.
    # @T        -> Temperature of the lithium surface evaporating lithium [K]  (float)
    # @L        -> list containing the Dimensions of the simulation domain [m] (list of floats)
    # @afac     -> Accuracy factor used to divide the mfp and mft [-]          (integer), default value = 3
    # @OptNC    -> Optimal number of simulation particles in a cell [#]        (integer), default value = 5
    # @mode     -> Simulation mode, axi or normal                              (string)
    # @printset -> If true print the settings, otherwise return the settings   (boolean)
    # @return   -> dictionary of settings consisting out of (key specified):
    #              - T    : Temperature of vapor cloud, echo of input          (float)
    #              - Nx   : Number of grid cells in the x direction            (integer)
    #              - Ny   : Number of grid cells in the y direction            (integer)
    #              - fnum : Simulation to real particle ratio                  (float)
    #              - dt   : time step of the simulation                        (float)
    #              - nrho : Vapor cloud density accroding to temperature       (float)
    #              - Ninit: Time steps taken to acheive steady state           (integer)
    #__________________________________________________________________________
    
    def DSMClimits(self,T,L,xmax=1,afac=3,optNC=5,mode="axi",printset=True):
        
        seed = rnd.randint(10000,99999)
        
        nrho = self.Density(T)
        
        Nmin = [int(np.ceil(Ldir/xmax)) for Ldir in L]
        Ngrid = []
        for Ldir in L:
            index = L.index(Ldir)
            Ndir = int(np.ceil(Ldir/(self.mean_free_path(nrho, T)/afac)))
            if Ndir < Nmin[index]:
                Ndir = Nmin[index]
            Ngrid.append(Ndir) 
        
        fnum = (L[0]/Ngrid[0])*(L[1]/Ngrid[1])*nrho/optNC
        if mode == "axi":
            fnum = fnum/(L[1]*(1-1/Ngrid[1]/2))
        
        dt = self.mean_free_time(nrho, T)/afac
        
        speed = self.u+self.thermal_speed(T)
        Ninit = int(100*np.ceil(np.max(L)/speed/dt/10))
        
        settings = {"seed"  : seed,
                    "Nx"    : Ngrid[0],
                    "Ny"    : Ngrid[1],
                    "fnum"  : fnum,
                    "dt"    : dt,
                    "T"     : T,
                    "Ninit" : Ninit,
                    "nrho"  : nrho}
        
        if printset:        # Display the settings
            [print('%s : %.7E'% (key,value)) for key, value in settings.items()]
        else:
            return settings
#______________________________________________________________________________   

class Lithium(DSMCSpecies):
    
    # Constructor -------------------------------------------------------------
    # Create an object based on the DSMC simulations species that contains
    # properties and functions that describe atomic lithium (LiHybSatNar) as a 
    # Variable soft sphere molecule, as calculated and determined by J. Schwartz
    # in his Thesis (2020), for use with SPARTA DSMC simulations.
    #__________________________________________________________________________
    
    def __init__(self):
        self.molmass = 6.941
        self.gamma = 5/3
        self.HeatVap = 145.92e6/self.molmass
        super().__init__("Atomic Lithium", self.molmass*u, 4.9996e-10, 0.740, 850, 1.0)
        

    # LiVaporPressure ---------------------------------------------------------
    # calculates the vapor pressure [Pa] of atomic lithium for a specified
    # temperature according to the relation of Browing and Potter, 1985.
    # @T     -> temperature of the vapor-liquid surface [K]                    (float)
    # return -> Vapor pressure of liquid lithium at specified T                (float)
    #__________________________________________________________________________
    
    def VaporPressure(self,T):
        A = 13.0719
        B = 18880.659
        C = 0.4942
        return 1e6*np.exp(A-B/T-C*np.log(T))

    # LiDensity ---------------------------------------------------------------
    # Calculate the vapor density of lithium vapor associated with the vapor 
    # pressure of saturated liquid lithium at a specified temperature.
    # @T     -> Temperature of the liquid and vapor cloud [K]                  (float)
    # return -> particle density of the vapor cloud [m^-3]                     (float)
    #__________________________________________________________________________
    
    def Density(self,T):
        p = self.VaporPressure(T)
        return p/kB/T
    
    # EvaporationFlux ---------------------------------------------------------
    # Calculate the mass flux from the liquid to the vapor cloud according to the
    # Hertz-Knudsen-Langmuir Equation.
    # @T     -> Temperature of the liquid and vapor cloud [K]                  (float)
    # return -> flow of vapor from surface to the vapor cloud [kg/m2/s]        (float)
    #__________________________________________________________________________
    
    def EvaporationFlux(self,T):
        n = self.Density(T)
        C = n/np.sqrt(2*np.pi)
        return self.mass*C*np.sqrt(kB*T/self.mass)
    
    # HeatCapacity ---------------------------------------------------------
    # Calculate the heat capacity of lithium at a specified temperature.
    # @T     -> Temperature of the lithium [K]                                 (float)
    # return -> Heat capacity of the lithium [J/kg/K]                          (float)
    #__________________________________________________________________________
    
    def HeatCapacity(self,T):
        A = [169.5520, 32.46972, 26.00896, 23.33408]
        B = [-882.7110, -2.635975, 5.632375, -2.772423]
        C = [1977.438, -6.327128, -4.0123227, 0.767421]
        D = [-1487.312, 4.230359, 0.873686, -0.003595]
        E = [-1.609635, 0.005686, 0.344150, -0.035246]
    
        t = T/1000
        if T <= 453.69:
            index = 0
        elif T <= 700.0:
            index = 1
        elif T <= 1620.12:
            index = 2
        else:
            index = 3

        return  (A[index] + B[index]*t + C[index]*t**2 + D[index]*t**3 +E[index]/(t**2))/(self.molmass/1000);
#______________________________________________________________________________

class IdealGas(DSMCSpecies):
    
    # Constructor -------------------------------------------------------------
    # Create an object based on the DSMC simulations species that contains
    # properties and functions that describe an ideal gas as a 
    # Variable soft sphere molecule, as implemented in SPARTA for DSMC simulations.
    #__________________________________________________________________________
    
    def __init__(self,IGtype,p0,T,mode):
        specie = SpeciesDict[IGtype]
        self.gamma = 1.4
        self.molmass = specie[1]
        super().__init__(specie[0], specie[2], specie[10], specie[11], specie[12], specie[13])
        self.p0 = p0
        self.mode = mode
        if mode == "choked":
            self.u = self.vSound(T) 
            ccond = {'Tc' : self.criticalT(T),
                     'cs' : self.u}
            [print('%s : %.4E'% (key,value)) for key, value in ccond.items()]
            
    # Density -----------------------------------------------------------------
    #
    #__________________________________________________________________________
    
    def Density(self,Tres):
        if self.mode == "choked":
            Tc = self.criticalT(Tres)
            pc = self.p0*(2/(self.gamma+1))**(self.gamma/(self.gamma-1))
            return pc/kB/Tc
        else:
            return self.p0/kB/Tres
        
    # vSound ------------------------------------------------------------------
    #
    #__________________________________________________________________________
    
    def vSound(self,Tres):
        if self.mode == "choked":
            Tc = self.criticalT(Tres)
            return np.sqrt(self.gamma*kB*Tc/self.mass)
        else:
            return np.sqrt(self.gamma*kB*Tres/self.mass)
        
    # criticalT ---------------------------------------------------------------
    #
    #__________________________________________________________________________
    
    def criticalT(self,T):
        return 2*T/(self.gamma+1)        
#______________________________________________________________________________
        
# main code if run ------------------------------------------------------------
if __name__ == "__main__":
    Particle = Lithium.DSMClimits(923, [0.191,0.061], xmax=0.005)

