# -*- coding: utf-8 -*-
#______________________________________________________________________________
# General modules -------------------------------------------------------------
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D

# Own modules -----------------------------------------------------------------
#from podostemaceous.constants import kB,u
#______________________________________________________________________________

class Line():
    
    # Constructor -------------------------------------------------------------
    # Create a line instance to store the data of a line.
    # mesh surface or used as a surface for radiation transfer.
    # @p1    -> First point coordinates in the shape: [x,y] [m]                (2x1 list of floats)
    # @p2    -> second point coordinates in the shape: [x,y] [m]               (2x1 list of floats)
    #__________________________________________________________________________
    
    def __init__(self,p1,p2):
        self.p1 = np.array(p1)
        self.p2 = np.array(p2)
        self.dim = len(p1)
        self.mid = self.__Midpoint()
        self.mag = self.__Length()
        self.A = self.__Area()
        self.slope = (p2[1]-p1[1])/(p2[0]-p1[0])
        self.offset = p1[1]-self.slope*p1[0]
        
    # Shift ------------------------------------------------------------------
    # shift the line with a specified vector
    # @Vec  -> vector of shift                                                 (2x1 list of floats)
    #__________________________________________________________________________
     
    def Shift(self,Vec):
        self.p1 += np.array(Vec)
        self.p2 += np.array(Vec)
        self.mid += np.array(Vec)
        self.A = self.__Area()
        
    # Rotate ------------------------------------------------------------------
    # Rotate the line aournd the midpoint with an angle theta. 
    # @theta  -> angle (in rad) of rotation                                    (float)
    #__________________________________________________________________________
    
    def Rotate(self,theta):
        M = np.array([[np.cos(theta),-1*np.sin(theta)],[np.sin(theta),np.cos(theta)]])
        Vec = self.p1 - self.mid
        nVec = np.array([M[0][0]*Vec[0]+M[0][1]*Vec[1],M[1][0]*Vec[0]+M[1][1]*Vec[1]])
        self.p1 = self.mid+nVec
        self.p2 = self.mid-nVec
        self.mid = self.__Midpoint()
        self.A = self.__Area()
        
    # Scale------------------------------------------------------------------
    #  scale the line with a specified factor.
    # @factorc  -> factor to scale the line with                               (floats)
    #__________________________________________________________________________
     
    def Scale(self,factor):
        self.p1 = self.mid+factor*(self.p1-self.mid)
        self.p2 = self.mid+factor*(self.p2-self.mid)
        self.mag = factor*self.mag
    
    # midpoint (private) ------------------------------------------------------ 
    # determine the midpoint of the line based on the two points stored in the 
    # line instance.
    # @return -> midpoint coordinates of the line as [x,y]                     (2x1 list of floats)])
    #__________________________________________________________________________
    
    def __Midpoint(self):
        return np.array([(self.p2[0]+self.p1[0])/2,(self.p2[1]+self.p1[1])/2])
    
    # Length (private) -------------------------------------------------------
    # determine the length of the line isntance that is created.
    # @return -> length (or magnitude) of the line                             (float)
    #__________________________________________________________________________
    
    def __Length(self):
        return np.linalg.norm(np.array([self.p2[0]-self.p1[0],self.p2[1]-self.p1[1]]))
    
    # Area (private) ----------------------------------------------------------
    # Determine the area under the line according to the trapezoid formula. 
    # @return -> area defined by the line and the x axis                       (float)
    #__________________________________________________________________________
    
    def __Area(self):
        return 0.5*(self.p1[1]+self.p2[1])*(self.p1[0]-self.p2[0])
    
    
#______________________________________________________________________________

class Surface(Line):
    
    # Constructor -------------------------------------------------------------
    # Create a surface instance to store the data of a Surface to be used in SPARTA meshes.
    # @gname -> Identifer for the group or par this surface is a part of          (string)
    # @stype -> Identifer for the surface type of this surface for use in the sim (integer)
    # @p1    -> First point coordinates in the shape: [x,y] [m]                (2x1 list of floats)
    # @p2    -> second point coordinates in the shape: [x,y] [m]               (2x1 list of floats)
    #__________________________________________________________________________
    
    def __init__(self,gname,stype,p1,p2):
        super().__init__(p1,p2)
        self.group = gname
        self.stype = stype
        self.normal = self.__NormVec()[:2]/np.linalg.norm(self.__NormVec())
        
    # Rotate (Override) -------------------------------------------------------
    # Rotate the surface aournd the midpoint with an angle thetaa nd update the normal vector 
    # @theta  -> angle (in rad) of rotation                                    (float)
    #__________________________________________________________________________
    
    def Rotate(self,theta):
        super().Rotate(theta)
        self.normal = self.__NormVec()[:2]/np.linalg.norm(self.__NormVec())
        
    # NormVec (private) -------------------------------------------------------
    # determine the right hand rule normal vector of the surface based on the points
    # stored in the surface instance.
    # @return -> normal vector of the surface as [x,y]                            (2x1 list of floats)])
    #__________________________________________________________________________
    
    def __NormVec(self):
        return np.cross(np.array([0,0,1]),np.array([(self.p2[0]-self.p1[0]),(self.p2[1]-self.p1[1]),0]))
#______________________________________________________________________________

class Mesh():
    
    # Constructor -------------------------------------------------------------
    # create an instance of a mesh that contains surfaces for use in SPARTA
    # simulations as well as plotting and file creation capabilities.
    # @name -> name for the mesh, cna be used for the output filename          (string)
    # @mode -> Define the simulation mode: "Normal" or "Axi"                   (string)
    #__________________________________________________________________________
    
    def __init__(self,name,mode):
        self.name = name
        self.mode = mode
        self.lastpoint = []
        self.surfaces = []
        self.margin = 0.0005
        # Simulation Bounds
        self.xlim = [1e30,0.0]
        self.ylim = [1e30,0.0]
        self.groups = {}
        
    # Mesh Creation/Manipulation ----------------------------------------------
    # SetPoint ----------------------------------------------------------------
    # Define a point as the last created point. Attach function will make use of
    # this point when creating a surface.
    # @point -> coordinates of the point in the shape: [x,y] [m]               (2x1 list of floats)
    #__________________________________________________________________________
    
    def SetPoint(self,point):
        self.lastpoint = point
        
    # AddSurface -----------------------------------------------------------------
    # Define a surface based on its two end points and add it to the mesh so it can
    # be used as a simulation surface in SPARTA simulations. The Simulation bounds
    # Are also checked and updated if the surface increase the simulation domain.
    # @gname  -> Identifer for the group or part this surface is a part of         (string)
    # @point1 -> First point coordinates in the shape: [x,y] [m]               (2x1 list of floats)
    # @point2 -> second point coordinates in the shape: [x,y] [m]              (2x1 list of floats)
    # @stype  -> Identifer for the surface type of this surface                   (integer) 
    #__________________________________________________________________________
        
    def AddSurface(self,gname,point1,point2,stype):
        self.SetPoint(point2)
        # Update bounds of Simulation
        if min(point1[0],point2[0]) < self.xlim[0]:
            self.xlim[0] = min(point1[0],point2[0])
        if max(point1[0],point2[0]) > self.xlim[1]:
            self.xlim[1] = max(point1[0],point2[0])
            
        if min(point1[1],point2[1]) < self.ylim[0]:
            self.ylim[0] = min(point1[1],point2[1])
        if max(point1[1],point2[1]) > self.ylim[1]:
            self.ylim[1] = max(point1[1],point2[1]) 
            
        # update group listing
        if gname not in self.groups:
            self.groups[gname] = []
            
        # Add surface into instance and to group   
        self.surfaces.append(Surface(gname,stype,point1,point2))
        self.groups[gname].append(len(self.surfaces)-1)
    
    # AttachSurface --------------------------------------------------------------
    # Create a surface using the last known point in the instance and a newly defined 
    # point for use as a surface in a SPARTA mesh.
    # @gname  -> Identifer for the group or part this surface is a part of         (string)
    # @point2 -> new point coordinates in the shape: [x,y] [m]                 (2x1 array of floats)
    # @stype  -> Identifer for the surface type of this surface                   (integer) 
    #__________________________________________________________________________    
    
    def AttachSurface(self,gname,point2,stype):
        if bool(self.lastpoint):
            point1 = self.lastpoint
            self.AddSurface(gname,point1,point2,stype)
     
    # AddCircle ---------------------------------------------------------------
    # Create an approximate circle composed of many small surface instances via
    # the AttachSurface method. the circle is based on an midpoint, radius, its
    # angular span and a startpoint.
    # @gname      -> Identifer for the group or part this circle is a part of  (string)
    # @midpoint   -> midpoint coordinates in the shape: [x,y] [m]              (2x1 list of floats)
    # @radius     -> Radius of the circle [m]                                  (float)
    # @theta      -> Angular span of the circle [rad]                          (float)
    # @Np         -> Number of piece that will make up the circle              (integer)
    # @stype      -> Identifer for the surface type of this surface               (integer)
    # @sign       -> 1 is counterclockwise creation, -1 is clockwise creation  (1 or -1) default = 1
    # @startpoint -> Starting point coordinates in the shape [x,y] [m]         (2x1 list of floats) default = point to the right of the midpoint
    #                ! Condition: make sure the starting point is actually on
    #                the circle defined by midpoint and radius.
    #__________________________________________________________________________    
    
    def AddCircle(self,gname,midpoint,radius,theta,Np,stype,sign=1,startpoint=[]):
        if not bool(startpoint):
            startpoint = [midpoint[0]+radius,midpoint[1]]
            shift = 0.0
        else:
            dis = np.linalg.norm(np.array([midpoint[0]+radius,midpoint[1]])-np.array(startpoint))
            shift = 2*np.arcsin(0.5*dis/radius)
            
        self.SetPoint(startpoint)
        
        for i in range(Np):
            angle = sign*(i+1)*theta/Np+shift
            pointx = np.cos(angle)*radius+midpoint[0]
            pointy = np.sin(angle)*radius+midpoint[1]
            point = [pointx,pointy]
            self.AttachSurface(gname, point, stype)
    
    # AttachCircle ------------------------------------------------------------
    # Create A circle that is attached to the last known point of the mesh by
    # using this as the starting point. See AddCircle for more info.
    # @gname      -> Identifer for the group or part this circle is a part of  (string)
    # @midpoint   -> midpoint coordinates in the shape: [x,y] [m]              (2x1 list of floats)
    # @radius     -> Radius of the circle [m]                                  (float)
    # @theta      -> Angular span of the circle [rad]                          (float)
    # @Np         -> Number of piece that will make up the circle              (integer)
    # @stype      -> Identifer for the surface type of this surface               (integer)
    #__________________________________________________________________________    
    
    def AttachCircle(self,gname,midpoint,radius,theta,Np,stype,sign=1):
        if bool(self.lastpoint):
            spoint = self.lastpoint
            self.AddCircle(gname,midpoint, radius, theta, Np, stype, sign=sign, startpoint=spoint)
            
    # RotateSurface --------------------------------------------------------------
    # Rotate an existing surface around a given point by a specified angle.
    # @ID    -> Index of he surface to rotate                                     (integer)
    # @point -> Point around which the rotation needs to occur                 (2x1 list of floats)
    # @theta -> Angle (in rad) of the rotation                                 (float)
    #__________________________________________________________________________        
    
    def RotateSurface(self,ID,point,theta):
        M = np.array([[np.cos(theta),-1*np.sin(theta)],[np.sin(theta),np.cos(theta)]])
        Vec = self.surfaces[ID].mid-np.array(point)
        nVec = np.array([M[0][0]*Vec[0]+M[0][1]*Vec[1],M[1][0]*Vec[0]+M[1][1]*Vec[1]])
        self.surfaces[ID].Shift(nVec-Vec)
        self.surfaces[ID].Rotate(theta)
        
    # gGeometry ---------------------------------------------------------------
    # for every defined group, determine the surface area (magnitude of the surfaces)
    # and the internal area enclosed by the group of surfaces. Make sure a group
    # consists of a closed system of surfaces.
    #__________________________________________________________________________
    
    def gGeometry(self):
        
        self.gGeo = {}
        
        for key,idxs in self.groups.items():
            self.gGeo[key] = np.zeros((2,))
            for idx in idxs:
                self.gGeo[key][0] += self.surfaces[idx].mag
                self.gGeo[key][1] += self.surfaces[idx].A
                
            self.gGeo[key][1] = abs(self.gGeo[key][1])
            
    # Segmenter ---------------------------------------------------------------
    # Segment an exisiting surface into a specified amount of surfaces.
    # @surface  -> Object of the surface class that needs to be segmented in surfaces.   (surface object)
    # @N     -> Number of surfaces to create from original                         (integer)
    # return -> N-size array filled with surface objects that span the original 
    #           surface length between the original points.                       (Nx1 array of surface objects)
    #__________________________________________________________________________
    
    def Segmenter(self,surface,N):
        surfaces = []
        p1 = np.array(surface.p1)
        p2 = np.array(surface.p2)
        for i in range(N):
            point1 = p1+i*(p2-p1)/N
            point2 = p1+(i+1)*(p2-p1)/N
            surfacepart = Surface(surface.group,surface.stype,point1,point2)
            surfacepart.pID = self.surfaces.index(surface)
            surfaces.append(surfacepart)
        return surfaces
            
    # SegmentSurface -------------------------------------------------------------
    # This function segments a currenlty existing surface into multiple surfaces.
    # @ID -> ID for the surface to segment                                        (integer)
    # @N  -> number of surfaces to segment the surface into                          (integer)
    #__________________________________________________________________________
    
    def SegmentSurface(self,ID,N):
        surface = self.surfaces[ID]
        list1 = self.surfaces[:ID]
        if ID != len(self.surfaces)-1:
            list2 = self.surfaces[ID+1:]
        else:
            list2 = []
        
        segment = self.Segmenter(surface,N)
        self.surfaces = list1 + segment + list2
        
        # update grouping
        for key,ilist in self.groups.items():
            for item in ilist:
                if item > ID:
                    self.groups[key][self.groups[key].index(item)] += N-1
                    
        index = self.groups[surface.group].index(ID)
        self.groups[surface.group].pop(index)
        [self.groups[surface.group].insert(index+i,ID+i) for i in range(N)]
        
    # AssignGrid --------------------------------------------------------------
    # Sort defined surfaces into an equidistant 2D grid of tiles. Each tile storing
    # the indexes of the surface that lie within it.
    # @size -> size of the equidistant 2D grid                                 (float)
    #__________________________________________________________________________
    
    def AssignGrid(self,size):
        if hasattr(self, 'surfaces'):
            # Create grid
            if not hasattr(self, 'bounds'):
                self.DetBounds(self.margin)
                
            Nx = int(np.ceil((self.bounds[1]-self.bounds[0])/size))
            Ny = int(np.ceil((self.bounds[3]-self.bounds[2])/size))
            self.dx = (self.bounds[1]-self.bounds[0])/Nx
            self.dy = (self.bounds[3]-self.bounds[2])/Ny
            
            step = 0.5*min(self.dx,self.dy)
            
            # Initalization
            self.Grid = [[[] for j in range(Ny)] for i in range(Nx)]
            
            # Sort surfaces into grid
            for surface in self.surfaces:
                lvec = np.array(surface.p2)-np.array(surface.p1)
                Nsteps = int(np.ceil(np.linalg.norm(lvec)/step))
                for N in range(Nsteps+1):
                    point = np.array(surface.p1)+(N/Nsteps)*lvec
                    ix = int((point[0]-self.bounds[0])//self.dx)
                    iy = int((point[1]-self.bounds[2])//self.dy)
                    if self.surfaces.index(surface) not in self.Grid[ix][iy]:
                        self.Grid[ix][iy].append(self.surfaces.index(surface))
                        
        else:
             print("No surfaces created yet, assigning is impossible")               
        
    # DetBounds ---------------------------------------------------------------
    # Determine the bounds of the simulation domain by add a margin to the stored
    # minimal and maximal values in the instance. If the simulation is axisymmetric,
    # the margin is not added to the lower y bound.
    # @margin -> margin around the minimal/maximal values to create a domain   (float)
    #__________________________________________________________________________
        
    def DetBounds(self,margin):
        self.bounds = [self.xlim[0]-margin,self.xlim[1]+margin,self.ylim[0]-margin,self.ylim[1]+margin]
        if self.mode == "Axi":
            self.bounds[2] += margin
            
    # Mesh Plotting -----------------------------------------------------------
    # PlotMesh ----------------------------------------------------------------
    # Plot the ,esh currently stored in the instance.
    # @figsize -> Size of the figure hte mesh is plotted in                    (float) default=6
    # @legend  -> if specified, uses a custom legend to display in the figure  (array of plt objects)
    # @colors  -> if specified, plot the mesh with color based on stype        (dictionary with stype keys)
    # @margin  -> if specified, change margin from default value               (float)
    #__________________________________________________________________________
    
    def PlotMesh(self,figsize=12,legend=[],colors=[],margin=[],qp='on',gp='on'):
        if not hasattr(self, 'Grid'):
            if not hasattr(self, 'bounds'):
                self.DetBounds(self.margin)
            self.AssignGrid(min(self.bounds[1]-self.bounds[0],self.bounds[3]-self.bounds[2])/50)
            
        fig = plt.figure(figsize=[figsize,(self.bounds[3]-self.bounds[2])/(self.bounds[1]-self.bounds[0])*figsize])
        ax = fig.add_subplot(111)
        if gp == 'on':
            ax.imshow(np.transpose(np.array([[int(bool((self.Grid[i][j]))) for j in range(len(self.Grid[0]))] for i in range(len(self.Grid))])),
                      cmap='Greys',alpha=0.1,origin='lower',extent=self.bounds)
        
        for surface in self.surfaces:
            xpoints = [surface.p1[0],surface.p2[0]]
            ypoints = [surface.p1[1],surface.p2[1]]
           
            if bool(colors):
                ax.plot(xpoints,ypoints,color=colors[surface.stype])
            else:
                ax.plot(xpoints,ypoints)
                
            if qp == 'on':
                ax.quiver(*surface.mid,*surface.normal,scale=None)
        
                
        if self.mode == "Axi":
            self.__Axiline(self.bounds[:2]) 
                
        if bool(legend):
            plt.legend(handles=legend,loc='best')
            
        plt.xlim(self.bounds[0],self.bounds[1])
        plt.ylim(self.bounds[2],self.bounds[3])
        plt.axis('equal')
        
        
        plt.xlabel('X-axis [m]',fontsize=15)
        plt.ylabel('Y-axis [m]',fontsize=15)
        
    
    
    # LineInter (private) --------------------------------------------------
    #
    #__________________________________________________________________________
    
    def __LineInter(self,p1,p2,p3,p4):
        tn = (p1[0]-p3[0])*(p3[1]-p4[1])-(p1[1]-p3[1])*(p3[0]-p4[0])
        td = (p1[0]-p2[0])*(p3[1]-p4[1])-(p1[1]-p2[1])*(p3[0]-p4[0])
        un = (p1[0]-p3[0])*(p1[1]-p2[1])-(p1[1]-p3[1])*(p1[0]-p2[0])
        ud = (p1[0]-p2[0])*(p3[1]-p4[1])-(p1[1]-p2[1])*(p3[0]-p4[0])
        if tn > 1e-10 and un > 1e-10  and tn < td and un < ud:
            return True
        else:
            return False    
    
    # Axiline (private) -------------------------------------------------------
    # plot a dashed line on y=0 to show that the simulation is axisymetric.
    # @limits -> bounds of the simulation domain in x direction                (2x1 list of floats)
    #__________________________________________________________________________
           
    def __Axiline(self,limits):
        plt.plot([limits[0],limits[1]],[0.0,0.0],linestyle='dashed',color='black')
        
    # File creation -----------------------------------------------------------
    # writefile ---------------------------------------------------------------
    # create a mesh file for SPARTA simulations using the surfaces stored in the
    # instance. File is written according to the SPARTA definitions. For ease of
    # use the number of surfaces and the bounds of the simulation are written into 
    # the top comment of the file.
    # @filename -> filename for the input file, default is the mesh name       (string)
    #__________________________________________________________________________

    def writefile(self,filename=[]):
        if 'bounds' not in dir(self):
            self.DetBounds(self.margin)
            
        if not bool(filename):
            filename=self.name

        file = open((filename + ".data"),'w')
        file.write("SPARTA Surface file. %d surfaces included. bounds are: %.6f %.6f %.6f %.6f\n" % (len(self.surfaces),*self.bounds))
        file.write("""
%d surfaces

Lines
\n""" % (len(self.surfaces)))
        for surface in self.surfaces:
            file.write("%d %d %.6f %.6f %.6f %.6f\n" % (self.surfaces.index(surface)+1,surface.stype,*surface.p1,*surface.p2))
#______________________________________________________________________________

# main code if run ------------------------------------------------------------
if __name__ == "__main__":
    # Create and plot mesh for the TDS geometry
    TDSm = Mesh("TDSv3mesh","Axi")
    
   
    
    