# -*- coding: utf-8 -*-
#______________________________________________________________________________
# General modules -------------------------------------------------------------
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
from matplotlib.patches import Patch
#from scipy.optimize import curve_fit

# Own Modules -----------------------------------------------------------------
from podostemaceous.constants import u,kB
from SPARTAmeshing import Line
#______________________________________________________________________________

class Cell:
    
    # Constructor -------------------------------------------------------------
    #
    #__________________________________________________________________________
    
    def __init__ ( self, xi,yi, ID, xlo, ylo, xc, yc, xhi, yhi, vol):
        self.xi         = xi
        self.yi         = yi
        self.ID         = int(ID)
        self.midpoint   = np.array([xc,yc])
        self.Aavail     = vol
        self.vals       = {}
        self.vals['dx'] = xhi-xlo
        self.vals['dy'] = yhi-ylo
        self.vals['A']  = self.CellArea(ylo,yhi)        
        
    # CellArea ----------------------------------------------------------------
    #
    #__________________________________________________________________________
    
    def CellArea(self,ylo,yhi):
        return np.pi*(yhi**2-ylo**2)
    
    # SetState ----------------------------------------------------------------
    #
    #__________________________________________________________________________
    
    def SetState(self,state):
        self.state = state
    
    # AddDensity --------------------------------------------------------------
    #
    #__________________________________________________________________________
    
    def AddDensity(self,n,nrho,massrho):
        self.vals['N']    = n
        self.vals['n']    = nrho 
        self.vals['rho']  = massrho
        
    # AddTemp -----------------------------------------------------------------
    #
    #__________________________________________________________________________
    
    def AddTemp(self,T):
        self.vals['T'] = T
        
    # AddPres -----------------------------------------------------------------
    #
    #__________________________________________________________________________
    
    def AddPres(self,p):
        self.vals['p']  = p
        
    # AddVel ---- -------------------------------------------------------------
    #
    #__________________________________________________________________________
    
    def AddVel(self,vx,vy,vz,pxrho,pyrho,pzrho):
        self.vals['vel']   = np.array([vx,vy,vz,np.sqrt(vx**2+vy**2+vz**2)])
        self.vals['prho']  = np.array([pxrho,pyrho,pzrho,np.sqrt(pxrho**2+pyrho**2+pzrho**2)])
        
    # CMach -------------------------------------------------------------------
    #
    #__________________________________________________________________________
    
    def CMach(self,molmass):
        self.vals['M'] = 0
        if self.vals['T'] > 1:
            c_s = np.sqrt((5.0/3.0)*kB*self.vals['T']/(molmass*u))
            self.vals['M'] = self.vals['vel'][3]/c_s
            
        if self.vals['M'] > 10:
                self.vals['M'] = 10
                
    # MeanFreePath ------------------------------------------------------------
    #
    #__________________________________________________________________________
     
    def MeanFreePath(self,mfp,L=0):
         self.vals['mfp'] = mfp
         if bool(L):       
             self.vals['Kn'] = mfp/L
         else:
             self.vals['Kn'] = None
#______________________________________________________________________________

class SpartaOutput:
    
    # Constructor -------------------------------------------------------------
    #
    #__________________________________________________________________________
    
    def __init__ ( self, path, simname, Species, plasmaloc=0.0):
        # files
        self.path = path
        self.simname = simname
        infile    = path+"in."+simname
        gridfile  = path+simname+"grid.out"
        valfile   = path+simname+"Gvals.out"
        
        # General
        self.Header = 9
        
        # Figure properties
        self.imagewidth = 12
        
        # Geomtery ------------------------------------------------------------
        inf=open(infile,'r')
        inflines = [line.split() for line in inf.read().splitlines() if bool(line.strip())]
        
        meshfile = inflines[10][1]
        self.bounds = [float(inflines[5][i+1]) for i in range(len(inflines[5])-3)]
        self.plasmaloc = plasmaloc
        
        self.Nx = int(inflines[6][1])
        self.Ny = int(inflines[6][2])
        self.qf = np.round(self.Nx/50)
        self.Nqx = int(self.Nx//self.qf)
        self.Nqy = int(self.Ny//self.qf)
        
        self.NoC = self.Nx*self.Ny
        
        self.Species = Species
        
        self.fnum = float(inflines[8][2])
        
        # Init ----------------------------------------------------------------
        self.Lines  = {}
        self.Cells = {}
        self.valnames = {'N'    : ["Particles","[-]"],
                         'n'    : ["Particle Density","[ m-3]"],
                         'rho'  : ["Mass Density","[kg/m3]"],
                         'T'    : ["Temperature","[K]"],
                         'p'    : ["Pressure","[Pa]"],
                         'vel'  : ["Velocity","[m/s]"],
                         'prho' : ["Momentum Density","[kg/s•m2]"],
                         'M'    : ["Mach Number","[-]"],
                         'mf'   : ["Mass Flux","[kg/m2/s]"],
                         'mdot' : ["Mass Flow","[kg/s]"],
                         'A'    : ["Cell Flow Area","[m2]"],
                         'mfp'  : ["Mean Free Path","[m]"],
                         'Kn'   : ["Knudsen Number","[-]"],
                         'vmag' : ["Velocity magnitude","[m/s]"],
                         'vx'   : ["x Velocity","[m/s]"]}
        self.X = np.zeros((self.Nx,self.Ny))
        self.Y = np.zeros((self.Nx,self.Ny))
        self.nozzle = np.zeros((self.Nx,self.Ny),dtype=bool)
        self.qX = np.zeros((self.Nqx,self.Nqy))
        self.qY = np.zeros((self.Nqx,self.Nqy))
        
        # Surfaces ------------------------------------------------------------
        mf = open(path+meshfile,'r')
        mflines = [line for line in mf.read().splitlines() if bool(line.strip())]
        self.LoadSurfaces(mflines)
        
        # Grid ----------------------------------------------------------------
        gf=open(gridfile,'r')
        gflines = [line.split() for line in gf.read().splitlines() if bool(line.strip())]
        gflines = gflines[self.Header:len(gflines)]
        self.LoadGrid(gflines)
        
        # Values --------------------------------------------------------------
        vf=open(valfile,'r')
        vflines = [line.split() for line in vf.read().splitlines() if bool(line.strip())]
        vflines = vflines[self.Header:len(vflines)]
        self.LoadValues(vflines)
    
# Functions for importing the SPARTA output    
    # LoadGrid ----------------------------------------------------------------
    #
    #--------------------------------------------------------------------------
    
    def LoadGrid(self,flines):
         for line in flines:
            ID = line[0]
            xi = (int(ID)-1)%self.Nx
            yi = (int(ID)-1)//self.Nx
            self.Cells[ID]=Cell(xi,yi,*[float(item) for item in line])
            self.X[xi][yi] = self.Cells[ID].midpoint[0]
            self.Y[xi][yi] = self.Cells[ID].midpoint[1]
            if xi%self.qf == 0 and yi%self.qf == 0:
                xqi = int(xi//self.qf)-1
                yqi = int(yi//self.qf)-1
                self.qX[xqi][yqi] = self.Cells[ID].midpoint[0]
                self.qY[xqi][yqi] = self.Cells[ID].midpoint[1]
                
    # LoadSurfaces ----------------------------------------------------------------
    #
    #--------------------------------------------------------------------------
    
    def LoadSurfaces(self,flines):
        if 'Points' in flines:
            self.points = {}
            NoP = int(flines[1].split()[0])
            pi = int(flines.index("Points"))+1
            
            for i in range(NoP):
                ID = flines[i+pi].split()[0]            
                self.points[ID] = [float(flines[i+pi].split()[1]),float(flines[i+pi].split()[2])]
                                   
            NoL = int(flines[2].split()[0])
            li = int(flines.index("Lines"))+1
                    
            for i in range(NoL):
                ID = flines[i+li].split()[0]
                IDp1 = flines[i+li].split()[2]
                IDp2 = flines[i+li].split()[3]
                self.Lines[ID] = Line("mesh", int(flines[i+li].split()[1]), self.points[IDp1], self.points[IDp2])
                
            self.__DetExt()
                
        elif 'Lines' in flines:
            NoL = int(flines[1].split()[0])
            li = int(flines.index("Lines"))+1
            
            for i in range(NoL):
                ID = flines[i+li].split()[0]
                p1 = [float(flines[i+li].split()[2]),float(flines[i+li].split()[3])]
                p2 = [float(flines[i+li].split()[4]),float(flines[i+li].split()[5])]
                self.Lines[ID] = Line("mesh", int(flines[i+li].split()[1]), p1, p2)
              
            self.__DetExt()
            
        else:
            print("Meshfile not in correct format")
            
    # LoadValues --------------------------------------------------------------
    #
    #--------------------------------------------------------------------------
    
    def LoadValues(self,flines):
        for line in flines:
            line = [float(item) for item in line]
            # Parameters
            ID = str(int(line[0]))
            
            # Values
            self.Cells[ID].AddDensity(line[1],line[2],line[3])
            self.Cells[ID].AddTemp(line[4])
            self.Cells[ID].AddPres(line[5])
            self.Cells[ID].AddVel(line[6],line[7],line[8],line[9],line[10],line[11])
            self.Cells[ID].CMach(self.Species.molmass)
            
            self.Cells[ID].SetState(self.__StateCell(self.Cells[ID]))
            
            # Knudsen
            if self.Cells[ID].vals['n'] > 0.0 and self.Cells[ID].vals['T'] > 0.0:
                mfp = self.Species.mean_free_path(self.Cells[ID].vals['n'],self.Cells[ID].vals['T'])
                L = self.__findr(self.Cells[ID])
                if bool(L):
                    L = 2*L
                self.Cells[ID].MeanFreePath(mfp,L=L)
            else:
                self.Cells[ID].vals['mfp'] = None
    
# Accessing Values ------------------------------------------------------------
    # getxline ----------------------------------------------------------------
    #
    # yloc -> [cm], y location where the line along the x-axis should be taken.
    #--------------------------------------------------------------------------
    
    def getxline(self,value,yloc,xlim=[],avg=False,pdir=""):
        
        yindex = int((yloc-self.bounds[2])//((self.bounds[3]-self.bounds[2])/self.Ny))*self.Nx
        if bool(xlim):
            N = int((xlim[1]-self.bounds[0])//((self.bounds[1]-self.bounds[0])/self.Nx))
            shift = int((xlim[0]-self.bounds[0])//((self.bounds[1]-self.bounds[0])/self.Nx))
        else:
            N = self.Nx
            shift = 0
        
        # create arrays
        val = []
        for i in range(N-shift):
            index = yindex+i+shift
            if self.__getVal(index,'state'):
                if bool(pdir) or pdir == 0:
                    val.append(self.__getVal(index,value)[pdir])
                else:
                    val.append(self.__getVal(index,value))
            elif not avg:
                val.append(None)
                    
        if avg:
            val = np.mean(np.array(val))
        else:
            val = np.array(val)
        
        return val
    
    # getyline -------------------------------------------------------------
    #
    # xloc -> [cm], x location where the line along the y-axis should be taken.
    #--------------------------------------------------------------------------
    
    def getyline(self,value,xloc,ylim=[],avg=False,pdir=""):
        
        xindex = int((xloc-self.bounds[0])//((self.bounds[1]-self.bounds[0])/self.Nx))
        if bool(ylim):
            N = int((ylim[1]-self.bounds[2])//((self.bounds[3]-self.bounds[2])/self.Ny))
            shift = int((ylim[0]-self.bounds[2])//((self.bounds[3]-self.bounds[2])/self.Ny))
        else:
            N = self.Ny
            shift = 0
        
        # create arrays
        val = []
        for i in range(N-shift):
            index = xindex+(i+shift)*self.Nx
            if self.__getVal(index,'state'):
                if bool(pdir) or pdir == 0:
                    val.append(self.__getVal(index,value)[pdir])
                else:
                    val.append(self.__getVal(index,value))
            elif not avg:
                val.append(None)
                    
        if avg:
            val = np.mean(np.array(val))
        else:
            val = np.array(val)
        
        return val
    
    # getSurface --------------------------------------------------------------
    #
    #--------------------------------------------------------------------------
    
    def getSurface(self,value,xlim=[],ylim=[],avg=False,pdir=""):
        
        if bool(xlim):
            Nx = int((xlim[1]-self.bounds[0])//((self.bounds[1]-self.bounds[0])/self.Nx))
            shiftx = int((xlim[0]-self.bounds[0])//((self.bounds[1]-self.bounds[0])/self.Nx))
        else:
            Nx = self.Nx
            shiftx = 0
            
        if bool(ylim):
            Ny = int((ylim[1]-self.bounds[2])//((self.bounds[3]-self.bounds[2])/self.Ny))
            shifty = int((ylim[0]-self.bounds[2])//((self.bounds[3]-self.bounds[2])/self.Ny))
        else:
            Ny = self.Ny
            shifty = 0
            
        if avg:
            val = []
            for i in range(Nx-shiftx):
                for j in range(Ny-shifty):
                    xi = i+shiftx
                    yi = j+shifty
                    ID = yi*self.Nx+xi
                    if self.Cells[str(ID+1)].state:
                        if bool(pdir) or pdir == 0:
                            val.append(self.__getVal(ID,value)[pdir])
                        else:
                            val.append(self.__getVal(ID,value))
            return np.mean(val)
                        
        else:
            val = np.zeros((Nx-shiftx,Ny-shifty))
            for i in range(Nx-shiftx):
                for j in range(Ny-shifty):
                    xi = i+shiftx
                    yi = j+shifty
                    ID = yi*self.Nx+xi
                    if self.Cells[str(ID+1)].state:
                        if bool(pdir) or pdir == 0:
                            val[i][j] = self.__getVal(ID,value)[pdir]
                        else:
                            val[i][j] = self.__getVal(ID,value)     
                    else:
                        val[i][j] = None
        
            return val
    
# Plotting --------------------------------------------------------------------         
    # Plotval -----------------------------------------------------------------
    #
    #--------------------------------------------------------------------------
    
    def ContourPlot(self,value,log="off",rotate="off",DaM="off"):
        name,unit = self.valnames[value]
        val = self.getSurface(value)
        
        if log == "on":
            val[val <= 0] = 1e-30
            val = np.log10(val)
            name = "log10("+name+")"
        
        if rotate == "on":
            val = np.flip(list(reversed(list(zip(*val)))),0)
            X = np.flip(list(reversed(list(zip(*self.Y)))),0)
            Y = np.flip(list(reversed(list(zip(*self.X)))),0)
            bounds = [self.bounds[2],self.bounds[3],self.bounds[0],self.bounds[1]]
        else:
            X = self.X
            Y = self.Y
            bounds = self.bounds
            
        if DaM == "on":
            bounds = [-bounds[1],bounds[1],bounds[2],bounds[3]]
            val = np.vstack((np.flip(val,0),val))
            X = np.vstack((-1*np.flip(X,0),X))
            Y = np.vstack((np.flip(Y,0),Y))
        
        
        # Create Figure
        fig = plt.figure(figsize=(self.imagewidth,(self.imagewidth+1)*((bounds[3]-bounds[2])/(bounds[1]-bounds[0]))),dpi=300)
        ax = fig.add_subplot(111)
        
        # Plotting
        image = ax.imshow(np.transpose(val),origin='lower',extent=bounds)
        if value == 'M':
            levels = np.linspace(0.5,4.5,9)
            cp    = ax.contour(X, Y, val,levels,colors='k')
        elif log == "on":
            nmin = int(np.ceil(np.nanmin(val)))
            nmax = int(np.floor(np.nanmax(val)))
            levels = np.arange(nmin,nmax+1,1,dtype=int)
            cp    = ax.contour(X, Y, val,levels,colors='k')
        else:
            cp    = ax.contour(X, Y, val,colors='k')
        self.__AddGeometryinPlot(ax,rotate=rotate,DaM=DaM)
        cbar = fig.colorbar(image,ax=ax,orientation='horizontal',pad=0.11,shrink=0.85) 
        
        # Formatting
        #ax.clabel(cp,inline=True,fmt="%2.2f",fontsize=8)
        #ax.set_title(name+' Plot',fontsize=18)
        ax.set_xlabel('x [m]',fontsize=20)
        ax.set_ylabel('y [m]',fontsize=20)
        ax.set_xlim(bounds[0],bounds[1])
        ax.set_ylim(bounds[2],bounds[3])
        ax.set_aspect('equal', 'box')
        ax.tick_params(labelsize=16)
        
        cbar.set_label(label=name+" "+unit,fontsize=20)
        cbar.ax.tick_params(labelsize=16) 
        
        fig.savefig(self.path+name+".png")
        plt.close()
        
    # QuiverPLot --------------------------------------------------------------
    #
    #--------------------------------------------------------------------------
    
    def QuiverPlot(self,value,showM="off",log="off",rotate="off",DaM="off"):
        name,unit = self.valnames[value]
        val1 = np.zeros((self.Nqx,self.Nqy))
        val2 = np.zeros((self.Nqx,self.Nqy))
        mag  = np.zeros((self.Nqx,self.Nqy))
        
        for ID in range(self.NoC):
            xi = self.__getVal(ID,'xi')
            yi = self.__getVal(ID,'yi')
            if (xi%self.qf) == 0 and (yi%self.qf) == 0:
                xqi = int(xi//self.qf)-1
                yqi = int(yi//self.qf)-1
                if self.Cells[str(ID+1)].state:
                    val1[xqi][yqi],val2[xqi][yqi],z,mag[xqi][yqi] = self.__getVal(ID,value)
                    if mag[xqi][yqi] != 0:
                        val1[xqi][yqi] = val1[xqi][yqi]/mag[xqi][yqi]
                        val2[xqi][yqi] = val2[xqi][yqi]/mag[xqi][yqi]
                else: 
                    val1[xqi][yqi] = float('NaN')
                    val2[xqi][yqi] = float('NaN')
                    
        if log == "on":
            mag[mag <= 0] = 1e-30
            mag = np.log10(mag)
            
                    
        if rotate == "on":
            intval = val1
            val1 = np.flip(list(reversed(list(zip(*val2)))),0)
            val2 = np.flip(list(reversed(list(zip(*intval)))),0)
            mag = np.flip(list(reversed(list(zip(*mag)))),0)
            qX = np.flip(list(reversed(list(zip(*self.qY)))),0)
            qY = np.flip(list(reversed(list(zip(*self.qX)))),0)
            bounds = [self.bounds[2],self.bounds[3],self.bounds[0],self.bounds[1]]
        else:
            qX = self.qX
            qY = self.qY
            bounds = self.bounds
            
        if DaM == "on":
            bounds = [-bounds[1],bounds[1],bounds[2],bounds[3]]
            val1 = np.vstack((-1*np.flip(val1,0),val1))
            val2 = np.vstack((np.flip(val2,0),val2))
            mag = np.vstack((np.flip(mag,0),mag))
            qX = np.vstack((-1*np.flip(qX,0),qX))
            qY = np.vstack((np.flip(qY,0),qY))
        
        
        # Create Figure
        fig = plt.figure(figsize=(self.imagewidth,(self.imagewidth-1)*((bounds[3]-bounds[2])/(bounds[1]-bounds[0]))),dpi=300)
        ax = fig.add_subplot(111)
        
        # Plotting
        qp = ax.quiver(qX, qY, val1, val2, mag,cmap='jet')
        self.__AddGeometryinPlot(ax,rotate=rotate,DaM=DaM)
        if showM != "off":
            M = self.getSurface('M')
            if rotate == "on":
                M = np.flip(list(reversed(list(zip(*M)))),0)
                X = np.flip(list(reversed(list(zip(*self.Y)))),0)
                Y = np.flip(list(reversed(list(zip(*self.X)))),0)
                bounds = [self.bounds[2],self.bounds[3],self.bounds[0],self.bounds[1]]
            else:
                X = self.X
                Y = self.Y
                bounds = self.bounds
                
            if DaM == "on":
                bounds = [-bounds[1],bounds[1],bounds[2],bounds[3]]
                M = np.vstack((np.flip(M,0),M))
                X = np.vstack((-1*np.flip(X,0),X))
                Y = np.vstack((np.flip(Y,0),Y))

                
            levels = np.linspace(0.5,4.5,9)
            cp = ax.contour(X, Y, M,levels,colors='k')
            ax.clabel(cp,inline=True,fontsize=8)
        
        cbar = fig.colorbar(qp,ax=ax,orientation='horizontal',pad=0.11,shrink=0.85)
            
        # Formatting
        ax.set_title(name+' Plot',fontsize=18)
        ax.set_xlabel('x [m]',fontsize=15)
        ax.set_ylabel('y [m]',fontsize=15)
        ax.set_xlim(bounds[0],bounds[1])
        ax.set_ylim(bounds[2],bounds[3])
        ax.set_aspect('equal', 'box')
        ax.tick_params(labelsize=12)
        
        cbar.set_label(label=name+" "+unit,fontsize=15)
        cbar.ax.tick_params(labelsize=12)
        
        fig.savefig(self.path+"q"+name+".png")
        plt.close()
            
    # Xplot -------------------------------------------------------------------
    #
    #--------------------------------------------------------------------------
    
    def Xplot(self,value,loc=[], yscale='linear'):
        name,unit = self.valnames[value]
        
        # values
        xplot = self.X[:,0]
        if bool(loc):
            val = self.getxline(value,loc)
        else:
            val = self.getXprofile(value)
        
        # Create Figure
        fig = plt.figure(figsize=(self.imagewidth,self.imagewidth*self.Ny/self.Nx+3),dpi=300)
        ax = fig.add_subplot(111)
        if yscale == 'log':
            plt.yscale("log")        
        
        # Plotting
        ax.plot(xplot,val,'k')
        if value == 'Kn':
            self.KnudsenLines(ax)
        
        # Formatting
        ax.set_title(name+' X-plot',fontsize=18)
        ax.set_xlabel('x [m]',fontsize=15)
        ax.set_ylabel(name+" "+unit,fontsize=15)
        ax.set_xlim(self.bounds[0],self.bounds[1])
        ax.grid()
        ax.tick_params(labelsize=12)
        
        fig.savefig(self.path+"x"+name+".png")
        plt.close()
        
    # CenterlinePlot -------------------------------------------------------------------
    #
    #--------------------------------------------------------------------------
    
    def CenterlinePlot(self,value,yscale='linear'):
        name,unit = self.valnames[value]
        
        # values
        xplot = self.X[:,0]
        val = self.getxline(value,self.plasmaloc)
        
        # Create Figure
        fig = plt.figure(figsize=(self.imagewidth*2,self.imagewidth),dpi=300)
        ax = fig.add_subplot(111)
        if yscale == 'log':
            plt.yscale("log")        
        
        # Plotting
        ax.plot(xplot,val,'k')
        if value == 'Kn':
            self.KnudsenLines(ax)
        
        # Formatting
        ax.set_title(name+' along y=%1.1f' % (self.plasmaloc),fontsize=18)
        ax.set_xlabel('x [m]',fontsize=15)
        ax.set_ylabel(name+" "+unit,fontsize=15)
        ax.set_xlim(self.bounds[0],self.bounds[1])
        ax.grid()
        ax.tick_params(labelsize=12)
        
        fig.savefig(self.path+"cl"+name+".png")
        plt.close()
        
    # Geometryplot -------------------------------------------------------------------
    #
    #--------------------------------------------------------------------------
    
    def Geometryplot(self,colors,lines_legend,Axi=False):
        if Axi:
            lines_legend.append(Line2D([0],[0],color='k',linestyle='dashed',label="Axisymmetry Axis"))
        
        # plot
        fig = plt.figure(figsize=(self.imagewidth,self.imagewidth*self.Ny/self.Nx+1),dpi=300)
        ax = fig.add_subplot(111)
        
        for key in self.Lines:
            x1,y1 = self.Lines[key].p1
            x2,y2 = self.Lines[key].p2
            
            ax.plot([x1,x2],[y1,y2],linewidth=2,color=colors[self.Lines[key].stype])
            
        if Axi:
            ax.plot([self.bounds[0],self.bounds[1]],[0.0,0.0],linewidth=2,color='k',linestyle='dashed')
            
        ax.set_xlabel('x [m]',fontsize=15)
        ax.set_ylabel('y [m]',fontsize=15)
        ax.set_xlim(self.bounds[0],self.bounds[1])
        ax.set_ylim(self.bounds[2],self.bounds[3])
        ax.set_aspect('equal')
        ax.tick_params(labelsize=12)
        ax.legend(handles=lines_legend,loc='best')
        
        fig.savefig(self.path+self.simname+"Geometry.png")
        plt.close()
        
# File manipulation -----------------------------------------------------------
    
    # CompileOutput -----------------------------------------------------------
    #
    #--------------------------------------------------------------------------
    
    def CompileOutput (self,Figures=True,rotate="off",DaM="off"):
        #  Figures
        if Figures:
            self.ContourPlot('N',log="on",rotate=rotate,DaM=DaM)
            self.ContourPlot('T',rotate=rotate,DaM=DaM)
            self.ContourPlot('n',log="on",rotate=rotate,DaM=DaM)
            self.ContourPlot('p',log="on",rotate=rotate,DaM=DaM)
            self.QuiverPlot('vel',showM="on",rotate=rotate,DaM=DaM)
            self.QuiverPlot('prho',log="on",rotate=rotate,DaM=DaM)
            self.ContourPlot('mfp',log="on",rotate=rotate,DaM=DaM)
            
            
# Internal Functions ----------------------------------------------------------
    # getVal ------------------------------------------------------------------
    #
    #--------------------------------------------------------------------------
    
    def __getVal(self,ID,val):
        ID  = str(ID+1)
        val = str(val)
        if val == 'xi':
            value = self.Cells[ID].xi
        elif val == 'yi':
            value = self.Cells[ID].yi
        elif val == "mf":
            value = self.Cells[ID].vals['prho'][0]    
        elif val == "x":
            value = self.Cells[ID].midpoint[0]
        elif val == "y":
            value = self.Cells[ID].midpoint[1]
        elif val == "state":
            value = self.Cells[ID].state
        elif val == "vmag":
            value = self.Cells[ID].vals["vel"][3]
        elif val == "vx":
            value = self.Cells[ID].vals["vel"][0]
        else:
            value = self.Cells[ID].vals[val]
        return value
    
    # StateCell ---------------------------------------------------------------
    #
    #--------------------------------------------------------------------------
            
    def __StateCell(self,cell):
        if cell.vals['N'] > 1E-30:
            return True
        else:
            return False
            
    # GeometryPLot ------------------------------------------------------------
    #
    #--------------------------------------------------------------------------
    
    def __AddGeometryinPlot(self,figure,rotate="off",DaM="off"):
        for key in self.Lines:
            if self.Lines[key].stype == 0:
                color = 'b'
            else:
                color = 'k'
                
            if rotate == "on":
                y1,x1 = self.Lines[key].p1
                y2,x2 = self.Lines[key].p2
            else:
                x1,y1 = self.Lines[key].p1
                x2,y2 = self.Lines[key].p2
            
            figure.plot([x1,x2],[y1,y2],linewidth=1.5,color=color)
            if DaM == "on":
                figure.plot([-x1,-x2],[y1,y2],linewidth=1.5,color=color)     
                
    # findr -------------------------------------------------------------------
    #
    #--------------------------------------------------------------------------
    
    def __findr(self,cell):
        xc = cell.midpoint[0]
        if xc < 0.0:
            xc = 0.0
        elif xc > self.xmax:
            xc = self.xmax
            
        for ID in self.Lines:
            p1 = np.array(self.Lines[ID].p1)
            p2 = np.array(self.Lines[ID].p2)
            if p1[0] != p2[0]:
                if xc >= p1[0] and xc <= p2[0]:
                    return ((xc-p1[0])/(p2[0]-p1[0])*(p2[1]-p1[1]))+p1[1]
                elif xc <= p1[0] and xc >= p2[0]:
                    return ((xc-p2[0])/(p1[0]-p2[0])*(p1[1]-p2[1]))+p2[1]
                
    # DetExt ------------------------------------------------------------------
    #
    #__________________________________________________________________________
    
    def __DetExt(self) : 
        self.xmin = 1e30
        self.xmax = 0.0
        self.ymin = 1e30
        self.ymax = 0.0
        
        for key,line in self.Lines.items():
            xpoints = [line.p1[0],line.p2[0]]
            if np.max(xpoints) > self.xmax:
                self.xmax = np.max(xpoints)
            if np.min(xpoints) < self.xmin:
                self.xmin = np.min(xpoints)
                
            ypoints = [line.p1[1],line.p2[1]]
            if np.max(ypoints) > self.ymax:
                self.ymax = np.max(ypoints)
            if np.min(ypoints) < self.ymin:
                self.ymin = np.min(ypoints)
                
    # KnudsenLines ------------------------------------------------------------
    #
    #--------------------------------------------------------------------------
    
    def KnudsenLines(self,ax):
        # Custom legend
        lines_legend = [Patch(facecolor='b', edgecolor='b',label="Continuum"),
                        Patch(facecolor='c', edgecolor='c',label="Slip"),
                        Patch(facecolor='g', edgecolor='g',label="Transition"),
                        Patch(facecolor='y', edgecolor='y',label="Free Molecular")]
        # Knudsen lines
        ax.axhspan(0,0.01,color='b',alpha=0.2)
        ax.axhspan(0.01,0.1,color='c',alpha=0.2)
        ax.axhspan(0.1,10,color='g',alpha=0.2)
        ax.axhspan(10.0,1000,color='y',alpha=0.2)
        ax.legend(handles=lines_legend,loc='best')
        ax.set_ylim(0.001,1000)
        
#______________________________________________________________________________